import sys, traceback
import time
import unittest
from pytoad import Connection
from groundhogday import GroundhogDay, RetryWithAirbrake, LINEAR

class MockPytoadConnection(Connection):
    
    exceptions = []
    
    def send_to_hoptoad(self, exception):
        self.exceptions.append( exception )

class AwesomeFakeClass():
    
    def __init__(self, testCase):
        self.last_exception = None
        self.executed_methods = {}
        self.iterations_before_notification = None
    
    @GroundhogDay(notification_threshold=2, sleep_time=0.5, maximum_retries=3, maximum_retry_callback='my_maximum_retry_callback', notification_callback='my_notification_callback', exception_callback='my_exception_callback')
    def function_that_raises_exception(self, add=-1):
        self.executed_methods['function_that_raises_exception'] = self.executed_methods.get('function_that_raises_exception', 0) + 1
        raise Exception('foobar exception')
    
    @GroundhogDay(backoff=LINEAR, sleep_time=0.3, maximum_retries=3)
    def linear_backoff_exception(self):
        raise Exception('my exception')
        
    @GroundhogDay(environment='test', backoff=LINEAR, sleep_time=50, maximum_retries=3)
    def test_environment_method(self):
        raise Exception('test exception')
        
    @RetryWithAirbrake(
        name='Python-Crawling-Stack',
        version='1.0.0',
        url='http://attachments.me',
        environment_name='production',
        api_key='[AIRBRAKE-API-KEY]',
        environment='test'
    )
    def function_that_sends_error_to_airbrake(self):
        raise Exception('an airbrake exception')
    
    @GroundhogDay
    def function_that_works(self):
        self.executed_methods['function_that_works'] = self.executed_methods.get('function_that_works', 0) + 1
    
    def my_maximum_retry_callback(self, e):
        self.last_exception = e
        self.executed_methods['my_maximum_retry_callback'] = self.executed_methods.get('my_maximum_retry_callback', 0) + 1
    
    def my_notification_callback(self, e):
        self.iterations_before_notification = self.executed_methods['function_that_raises_exception']
    
    def my_exception_callback(self, e):
        self.executed_methods['my_exception_callback'] = self.executed_methods.get('my_exception_callback', 0) + 1

class TestGroundhogDay(unittest.TestCase):

    def test_original_function_executed_with_no_args(self):
        global call_me
        call_me = False

        @GroundhogDay
        def f():
            global call_me
            call_me = True
        f()
        
        afc = AwesomeFakeClass(self)
        afc.function_that_works()
        
        self.assertEqual(call_me, True)
        self.assertEqual(afc.executed_methods['function_that_works'], 1)

    def test_original_function_executed_with_args(self):
        global value, value2
        value = None

        @GroundhogDay
        def f(a, v2=99):
            global value, value2
            value = a
            value2 = v2
        f(2)
        
        self.assertEqual(value, 2)
        self.assertEqual(value2, 99)

    def test_exeption_raised_after_maximum_retries(self):
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_raises_exception(add=1)
        except:
            pass
        self.assertEqual(3, afc.executed_methods['function_that_raises_exception'])
    
    def test_exponential_backoff_works(self):
        start = time.time()
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_raises_exception(add=1)
        except:
            pass
        self.assertTrue(time.time() - start >= 0.3)
        self.assertTrue(time.time() - start <= 0.5)
    
    def test_maximum_retry_callback_executed_with_appropriate_exception(self):
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_raises_exception(add=1)
        except:
            pass
        self.assertEqual('foobar exception', str(afc.last_exception))
        self.assertEqual(1, afc.executed_methods['my_maximum_retry_callback'])
        
    def test_notificadtion_callback_executed_after_appropriate_number_of_exceptions(self):
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_raises_exception(add=1)
        except:
            pass
        self.assertEqual(2, afc.iterations_before_notification)
    
    def test_linear_backoff(self):
        start = time.time()
        afc = AwesomeFakeClass(self)
        try:
            afc.linear_backoff_exception(add=1)
        except:
            pass
        self.assertTrue(time.time() - start >= 0.5)
        self.assertTrue(time.time() - start <= 0.7)
    
    def test_test_environment_should_not_sleep(self):
        start = time.time()
        afc = AwesomeFakeClass(self)
        try:
            afc.test_environment_method(add=1)
        except:
            pass
        self.assertTrue(time.time() - start < 0.5)
    
    def test_exception_callback(self):
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_raises_exception(add=1)
        except:
            pass
        self.assertEqual(3, afc.executed_methods['my_exception_callback'])
        
    def test_airbrake_retry_decorator(self):
        afc = AwesomeFakeClass(self)
        try:
            afc.function_that_sends_error_to_airbrake()
        except:
            pass



import types
import inspect
import functools
from pytoad import Connection
from groundhogday import GroundhogDayClass

class AirbrakeNotifier(object):
    
    def __init__(self, **kwargs):
        self.environment = kwargs.get('environment', 'production')
        self.additional_information = kwargs['additional_information']
        self.pytoad_connection = Connection(**kwargs)
        
    def __call__(self, e):
        if not self.environment == 'test':
            self.pytoad_connection.send_to_hoptoad(e, self.additional_information)
        else:
            print 'Would send (%s %s) to airbrake' % (e, self.additional_information)

def _get_additional_information(f):
    try:
        sourcelines = inspect.getsourcelines(f)
        return "method=%s file=%s line=%s" % (f.__name__, inspect.getfile(f).split('/')[-1],sourcelines[1])
    except:
        return None

def RetryWithAirbrake(*args, **kwargs):
    if len(args) and type(args[0]) == types.FunctionType: # We called without parenthesis.
        functools.wraps(args[0])
        return GroundhogDayClass(args[0], notification_callback=AirbrakeNotifier(additional_information=_get_additional_information(args[0])))()
    else:
        def wrap(f):
            functools.wraps(f)
            kwargs['additional_information'] = _get_additional_information(f)
            kwargs['notification_callback'] = AirbrakeNotifier(**kwargs)
            return GroundhogDayClass(f, **kwargs)()
        return wrap


import functools
from raven import Client
from groundhogday import GroundhogDayClass

class SentryNotifier(object):
    
    def __init__(self, **kwargs):
        self.environment = kwargs.get('environment', 'production')
        self.sentry_client = Client(**kwargs)
        
    def __call__(self, e):
        if not self.environment == 'test':
            self.sentry_client.get_ident(self.sentry_client.captureException())
        else:
            print 'Would send [%s] to sentry.' % e

def RetryWithSentry(*args, **kwargs):
    def wrap(f):
        functools.wraps(f)
        kwargs['notification_callback'] = SentryNotifier(**kwargs)
        return GroundhogDayClass(f, **kwargs)()
    return wrap


#!/usr/bin/env python
#from distutils.core import setup
from setuptools import setup, find_packages

setup(name="groundhogday",
      version="0.0.5",
      description="GroundhogDay lets you repeat an operation until you get it right.",
      author="Benjamin Coe",
      author_email="ben@attachments.me",
      entry_points = {},
      url="https://github.com/bcoe/groundhogday",
      packages = find_packages(),
      install_requires = ['Pytoad==1.0.7', 'raven'],
      tests_require=['nose']
)



import time
import types
import inspect
import functools

LINEAR = 0
EXPONENTIAL = 1

class GroundhogDayClass(object):

    def __init__(self, f, **kwargs):
        self.f = f
        self.maximum_retries = kwargs.get('maximum_retries', 4)
        self.sleep_time = kwargs.get('sleep_time', 1.7)
        self.backoff = kwargs.get('backoff', 1)
        self.notification_threshold = kwargs.get('notification_threshold', 3)
        self.exception_callback_name = kwargs.get('exception_callback', None)
        self.notification_callback_name = kwargs.get('notification_callback', None)
        self.maximum_retry_callback_name = kwargs.get('maximum_retry_callback', None)
        self.environment = kwargs.get('environment', 'production')
        
    def _sleep(self, retry_count):
        if self.environment == 'test':
            return
        
        if self.backoff == LINEAR:
            time.sleep(self.sleep_time)
        else:
            backoff = self.sleep_time
            for i in range(0, retry_count):
                backoff *= backoff
            time.sleep(backoff)
            
    def __call__(self):
        def wrap(*args, **kwargs):
            last_exception = None
            
            for retry_count in range(0, self.maximum_retries):
                if retry_count:
                    self._sleep(retry_count)
                    
                try:
                    return self.f(*args, **kwargs)
                except Exception, e:
                    last_exception = e
                    exception_callback = self._get_wrapped_method(args[0], self.exception_callback_name)
                    if exception_callback:
                        exception_callback(last_exception)
                        
                    notification_callback = self._get_wrapped_method(args[0], self.notification_callback_name)
                    if notification_callback and retry_count == (self.notification_threshold - 1):
                        notification_callback(last_exception)
            
            if self.maximum_retry_callback_name:
                maximum_retry_callback = self._get_wrapped_method(args[0], self.maximum_retry_callback_name)
                if maximum_retry_callback:
                    maximum_retry_callback(last_exception)
            
            raise last_exception
        return wrap
    
    def _get_wrapped_method(self, wrapped_self, method_name):
        if callable(method_name): # Allow for callbacks to be callables rather than method names.
            return method_name
            
        members = {}
        for kv in inspect.getmembers(wrapped_self, inspect.ismethod):
            if kv[0] == method_name:
                return kv[1]
            
def GroundhogDay(*args, **kwargs):
    if len(args) and type(args[0]) == types.FunctionType: # We called without parenthesis.
        functools.wraps(args[0])
        return GroundhogDayClass(args[0])()
    else:
        def wrap(f):
            functools.wraps(f)
            return GroundhogDayClass(f, **kwargs)()
        return wrap